==> ns-1.0b2.tar.gz: ns-1.0b2/link.cc <==
/*
 * Copyright (c) 1990-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: link.cc,v 1.21 96/01/26 14:33:04 mccanne Exp $ (LBL)";
#endif

#include <stdlib.h>
#include <math.h>
#include <sys/types.h>

#include "link.h"
#include "packet.h"
#include "node.h"
#include "object.h"
#include "trace.h"

void Queue::remove(Packet* target)
{
	for (Packet** p = &head_; *p != 0; p = &(*p)->next_) {
		Packet* pkt = *p;
		if (pkt == target) {
			*p = pkt->next_;
			if (tail_ == &pkt->next_)
				tail_ = p;
			--len_;
			return;
		}
	}
	abort();
}

Link::Link() : busy_(0), neighbor_(0), qnp_(0), qnb_(0), qtrace_(0)
{
	Tcl& tcl = Tcl::instance();
	link_bw("ns_link", "bandwidth", &bandwidth_, 0);
	link_time("ns_link", "delay", &delay_, 0);
	link_int("ns_link", "queue-limit", &qlim_, 0);

	link_int(0, "queue-length", &qnp_, 1);
	link_int(0, "queue-size", &qnb_, 1);

	ncs_ = 10;
	cs_ = new classStat[ncs_];
}

void Link::grow(int m)
{
	int i;
	int n = ncs_ << 1;
	while (n < m)
		n <<= 1;

	classStat* p = new classStat[n];
	for (i = 0; i < ncs_; ++i)
		p[i] = cs_[i];

	for (; i < n; ++i) {
		p[i].np = 0;
		p[i].nb = 0;
		p[i].nd = 0;
	}
	ncs_ = n;
	delete cs_;
	cs_ = p;
}

/*
 * Called every time the queue size changes.
 * We integrate the queue size and keep.
 */
void Link::update(const Packet* pkt, int how)
{
	int c = pkt->class_;
	check_class(c);
	classStat& cs = cs_[c];
	if (how == 'd') {
		/* drop */
		++cs.nd;
	} else if (how != 'a') {
		/* normal depature */
		++cs.np;
		cs.nb += pkt->size_;
	}

	/* sum up the queue size (in both packets and bytes) */
	cs.sb.newpt(qnb_);
	cs.sp.newpt(qnp_);
	/* integrate the queue size (in both packets and bytes) */
	double now = Scheduler::instance().clock();
	ib_.newpt(now, qnb_);
	ip_.newpt(now, qnp_);
	iu_.newpt(now, (qnp_ != 0) ? 1. : 0.);
}

void Link::send()
{
	Packet* pkt = deque();
	if (pkt != 0) {
		double txtime = pkt->size_ * 8. / bandwidth_;
		Scheduler& s = Scheduler::instance();
		pkt->lsrc_ = src_;
		pkt->ldst_ = dst_;
		s.schedule(neighbor_, pkt, txtime + delay_);
		s.schedule(this, &intr_, txtime);
		busy_ = 1;
		busytime_ += txtime;
		if (trace_ != 0)
			trace_->hop(src_, dst_, pkt);
	}
}

/*
 * The only way this handler is called is from 
 * a "transmission interrupt".
 */
void Link::handle(Event* e)
{
	/*XXX*/
	if (e != &intr_)
		abort();

	busy_ = 0;
	send();
}

/*
 * Accept a packet for transmission on this link.
 * The node class calls this method.
 */
void Link::send(Packet* pkt)
{
	enque(pkt);
	if (!busy_)
		send();
}

void Link::log_packet_arrival(Packet* p)
{
	++qnp_;
	qnb_ += p->size_;
	update(p, 'a');
	if (trace_ != 0)
		trace_->enque(src_, dst_, p);
	if (qtrace_ != 0)
		plot();
}

void Link::log_packet_departure(Packet* p, int how)
{
	qnb_ -= p->size_;
	--qnp_;
	update(p, how);
	if (trace_ != 0) {
		if (how == 'd')
			trace_->drop(src_, dst_, p);
		else 
			trace_->deque(src_, dst_, p);
	}
	if (qtrace_ != 0)
		plot();
}
	
void Link::plot()
{
	double now = Scheduler::instance().clock();
	Trace* t = qtrace_;
	sprintf(t->buffer(), "q %g %d", now, qnp_);
	t->dump();
	sprintf(t->buffer(), "b %g %d", now, qnb_);
	t->dump();
}

/*
 * $link insert $src $dst
 * $link stat $class $which
 * $link integral qsize
 * $link integral qlen
 * $link integral util
 * $link queue-trace $trace
 */
int Link::command(int argc, const char*const* argv)
{
	Tcl& tcl = Tcl::instance();
	if (argc == 2) {
		if (strcmp(argv[1], "reset") == 0) {
			reset();
			return (TCL_OK);
		}
	} else if (argc == 3) {
		if (strcmp(argv[1], "integral") == 0) {
			double now = Scheduler::instance().clock();
			if (strcmp(argv[2], "util") == 0) {
				iu_.update(now);
				tcl.resultf("%g", iu_.sum());
				return (TCL_OK);
			}
			if (strcmp(argv[2], "qsize") == 0) {
				ib_.update(now);
				tcl.resultf("%g", ib_.sum());
				return (TCL_OK);
			}
			if (strcmp(argv[2], "qlen") == 0) {
				ip_.update(now);
				tcl.resultf("%g", ip_.sum());
				return (TCL_OK);
			}
		}
		if (strcmp(argv[1], "queue-trace") == 0) {
			qtrace_ = (Trace*)TclObject::lookup(argv[2]);
			if (qtrace_ == 0) {
				Tcl& tcl = Tcl::instance();
				tcl.resultf("link queue-trace: no trace %s",
					    argv[2]);
				return (TCL_ERROR);
			}
			return (TCL_OK);
		}
	} else if (argc == 4) {
		if (strcmp(argv[1], "install") == 0) {
			/*XXX check for null */
			Node* src = (Node*)TclObject::lookup(argv[2]);
			Node* dst = (Node*)TclObject::lookup(argv[3]);
			if (src == dst) {
				Tcl& tcl = Tcl::instance();
				tcl.resultf("link install: self-loop");
				return (TCL_ERROR);
			}
			neighbor(dst);
			src->add_link(this);
			return (TCL_OK);
		}
		if (strcmp(argv[1], "stat") == 0) {
			int c = atoi(argv[2]);
			if (c >= ncs_) {
				tcl.result("0");
				return (TCL_OK);
			}
			classStat& cs = cs_[c];
			const char* s = argv[3];
			if (strcmp(s, "packets") == 0) {
				tcl.resultf("%d", cs.np);
				return (TCL_OK);
			}
			if (strcmp(s, "bytes") == 0) {
				tcl.resultf("%d", cs.nb);
				return (TCL_OK);
			}
			if (strcmp(s, "drops") == 0) {
				tcl.resultf("%d", cs.nd);
				return (TCL_OK);
			}
			/* average qlen seen by this class on arrival */
			if (strcmp(s, "mean-qlen") == 0) {
				tcl.resultf("%g", cs.sp.mean());
				return (TCL_OK);
			}
			/* average q size seen by this class on arrival */
			if (strcmp(s, "mean-qsize") == 0) {
				tcl.resultf("%g", cs.sb.mean());
				return (TCL_OK);
			}
		}
	}
	return (NsObject::command(argc, argv));
}

void Link::reset()
{
	for (int i = 0; i < ncs_; ++i) {
		cs_[i].np = 0;
		cs_[i].nb = 0;
		cs_[i].nd = 0;
		cs_[i].sb.reset();
		cs_[i].sp.reset();
	}
	double now = Scheduler::instance().clock();
	iu_.reset(now);
	ib_.reset(now);
	ip_.reset(now);
}

==> ns-1.1.tar.gz: ns-1.1/link.cc <==
/*
 * Copyright (c) 1990-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: link.cc,v 1.2 96/10/09 17:17:54 kfall Exp $ (LBL)";
#endif

#include <stdlib.h>
#include <math.h>
#include <sys/types.h>

#include "link.h"
#include "packet.h"
#include "node.h"
#include "object.h"
#include "trace.h"

void Queue::remove(Packet* target)
{
	for (Packet** p = &head_; *p != 0; p = &(*p)->next_) {
		Packet* pkt = *p;
		if (pkt == target) {
			*p = pkt->next_;
			if (tail_ == &pkt->next_)
				tail_ = p;
			--len_;
			return;
		}
	}
	abort();
}

Link::Link() : busy_(0), neighbor_(0), qnp_(0), qnb_(0), qtrace_(0),
    flowmgr_(0)
{
	Tcl& tcl = Tcl::instance();
	link_bw("ns_link", "bandwidth", &bandwidth_, 0);
	link_time("ns_link", "delay", &delay_, 0);
	link_int("ns_link", "queue-limit", &qlim_, 0);

	link_int(0, "queue-length", &qnp_, 1);
	link_int(0, "queue-size", &qnb_, 1);

	ncs_ = 10;
	cs_ = new classStat[ncs_];
}

void Link::grow(int m)
{
	int i;
	int n = ncs_ << 1;
	while (n < m)
		n <<= 1;

	classStat* p = new classStat[n];
	for (i = 0; i < ncs_; ++i)
		p[i] = cs_[i];

	for (; i < n; ++i) {
		p[i].np = 0;
		p[i].nb = 0;
		p[i].nd = 0;
	}
	ncs_ = n;
	delete cs_;
	cs_ = p;
}

/*
 * Called every time the queue size changes.
 * We integrate the queue size and keep.
 */
void Link::update(Packet *pkt, int how)
{
	int c = pkt->class_;
	check_class(c);
	classStat& cs = cs_[c];
	if (how == 'd') {
		/* drop */
		++cs.nd;
	} else if (how != 'a') {
		/* normal depature */
		++cs.np;
		cs.nb += pkt->size_;
		cs.sdel.newpt(Scheduler::instance().clock() - pkt->qtime_);
	} else if (how == 'a') {
		/* arrival */
 	        pkt->qtime_ = Scheduler::instance().clock();
        }
	/* sum up the queue size (in both packets and bytes) */
	cs.sb.newpt(qnb_);
	cs.sp.newpt(qnp_);
	/* integrate the queue size (in both packets and bytes) */
	double now = Scheduler::instance().clock();
	ib_.newpt(now, qnb_);
	ip_.newpt(now, qnp_);
	iu_.newpt(now, (qnp_ != 0) ? 1. : 0.);
}

void Link::send()
{
	Packet* pkt = deque();
	if (pkt != 0) {
		double txtime = pkt->size_ * 8. / bandwidth_;
		Scheduler& s = Scheduler::instance();
		pkt->lsrc_ = src_;
		pkt->ldst_ = dst_;
		s.schedule(neighbor_, pkt, txtime + delay_);
		s.schedule(this, &intr_, txtime);
		busy_ = 1;
		busytime_ += txtime;
		if (trace_ != 0)
			trace_->hop(src_, dst_, pkt);
	}
}

/*
 * The only way this handler is called is from 
 * a "transmission interrupt".
 */
void Link::handle(Event* e)
{
	/*XXX*/
	if (e != &intr_)
		abort();

	busy_ = 0;
	send();
}

/*
 * Accept a packet for transmission on this link.
 * The node class calls this method.
 */
void Link::send(Packet* pkt)
{
	enque(pkt);
	if (!busy_)
		send();
}

void Link::log_packet_arrival(Packet* p)
{
	++qnp_;
	qnb_ += p->size_;
	update(p, 'a');
	if (trace_ != 0)
		trace_->enque(src_, dst_, p);
	if (qtrace_ != 0)
		plot();
}

void Link::log_packet_departure(Packet* p, int how)
{
	qnb_ -= p->size_;
	--qnp_;
	update(p, how);
	if (trace_ != 0) {
		if (how == 'd')
			trace_->drop(src_, dst_, p);
		else {
			trace_->deque(src_, dst_, p);
		}
	}
	if (qtrace_ != 0)
		plot();
}
	
void Link::plot()
{
	double now = Scheduler::instance().clock();
	Trace* t = qtrace_;
	sprintf(t->buffer(), "q %g %d", now, qnp_);
	t->dump();
	sprintf(t->buffer(), "b %g %d", now, qnb_);
	t->dump();
}

/*
 * $link insert $src $dst
 * $link stat $class $which
 * $link integral qsize
 * $link integral qlen
 * $link integral util
 * $link queue-trace $trace
 */
int Link::command(int argc, const char*const* argv)
{
	Tcl& tcl = Tcl::instance();
	if (argc == 2) {
		if (strcmp(argv[1], "reset") == 0) {
			reset();
			return (TCL_OK);
		}
	} else if (argc == 3) {
		if (strcmp(argv[1], "integral") == 0) {
			double now = Scheduler::instance().clock();
			if (strcmp(argv[2], "util") == 0) {
				iu_.update(now);
				tcl.resultf("%g", iu_.sum());
				return (TCL_OK);
			}
			if (strcmp(argv[2], "qsize") == 0) {
				ib_.update(now);
				tcl.resultf("%g", ib_.sum());
				return (TCL_OK);
			}
			if (strcmp(argv[2], "qlen") == 0) {
				ip_.update(now);
				tcl.resultf("%g", ip_.sum());
				return (TCL_OK);
			}
		}
		if (strcmp(argv[1], "queue-trace") == 0) {
			qtrace_ = (Trace*)TclObject::lookup(argv[2]);
			if (qtrace_ == 0) {
				Tcl& tcl = Tcl::instance();
				tcl.resultf("link queue-trace: no trace %s",
					    argv[2]);
				return (TCL_ERROR);
			}
			return (TCL_OK);
		}
		if (strcmp(argv[1], "flow-mgr") == 0) {
			flowmgr_ = (FlowManager*)TclObject::lookup(argv[2]);
			if (flowmgr_ == 0) {
				Tcl& tcl = Tcl::instance();
				tcl.resultf("link flow-mgr: no flowmgr %s",
					    argv[2]);
				return (TCL_ERROR);
			}
			return (TCL_OK);
		}
	} else if (argc == 4) {
		if (strcmp(argv[1], "install") == 0) {
			/*XXX check for null */
			Node* src = (Node*)TclObject::lookup(argv[2]);
			Node* dst = (Node*)TclObject::lookup(argv[3]);
			if (src == dst) {
				Tcl& tcl = Tcl::instance();
				tcl.resultf("link install: self-loop");
				return (TCL_ERROR);
			}
			neighbor(dst);
			src->add_link(this);
			return (TCL_OK);
		}
		if (strcmp(argv[1], "stat") == 0) {
			int c = atoi(argv[2]);
			if (c >= ncs_) {
				tcl.result("0");
				return (TCL_OK);
			}
			classStat& cs = cs_[c];
			const char* s = argv[3];
			if (strcmp(s, "packets") == 0) {
				tcl.resultf("%d", cs.np);
				return (TCL_OK);
			}
			if (strcmp(s, "bytes") == 0) {
				tcl.resultf("%d", cs.nb);
				return (TCL_OK);
			}
			if (strcmp(s, "drops") == 0) {
				tcl.resultf("%d", cs.nd);
				return (TCL_OK);
			}
			/* average qlen seen by this class on arrival */
			if (strcmp(s, "mean-qlen") == 0) {
				tcl.resultf("%g", cs.sp.mean());
				return (TCL_OK);
			}
			/* average q size seen by this class on arrival */
			if (strcmp(s, "mean-qsize") == 0) {
				tcl.resultf("%g", cs.sb.mean());
				return (TCL_OK);
			}
			/* average q delay seen by this class on departure */
			if (strcmp(s, "mean-qdelay") == 0) {
				tcl.resultf("%g", cs.sdel.mean());
				return (TCL_OK);
			}
		}
	}
	return (NsObject::command(argc, argv));
}

void Link::reset()
{
	for (int i = 0; i < ncs_; ++i) {
		cs_[i].np = 0;
		cs_[i].nb = 0;
		cs_[i].nd = 0;
		cs_[i].sb.reset();
		cs_[i].sp.reset();
		cs_[i].sdel.reset();
	}
	double now = Scheduler::instance().clock();
	iu_.reset(now);
	ib_.reset(now);
	ip_.reset(now);
}
